------------------------------------------------------------------------------
--                                                                          --
--                GNU ADA RUN-TIME LIBRARY (GNARL) COMPONENTS               --
--                                                                          --
--                   S Y S T E M . O S _ I N T E R F A C E                  --
--                                                                          --
--                                  B o d y                                 --
--                                                                          --
--                              $Revision: 1.2 $
--                                                                          --
--           Copyright (C) 1999-2000 Free Software Foundation, Inc.         --
--                                                                          --
-- GNARL is free software; you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. GNARL is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNARL; see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
-- GNARL was developed by the GNARL team at Florida State University. It is --
-- now maintained by Ada Core Technologies Inc. in cooperation with Florida --
-- State University (http://www.gnat.com).                                  --
--                                                                          --
------------------------------------------------------------------------------

--  This is a UnixWare (Native) version of this package

pragma Polling (Off);
--  Turn off polling, we do not want ATC polling to take place during
--  tasking operations. It causes infinite loops and other problems.

with Interfaces.C;

package body System.OS_Interface is

   use Interfaces.C;

   -----------------
   -- To_Duration --
   -----------------

   function To_Duration (TS : timespec) return Duration is
   begin
      return Duration (TS.tv_sec) + Duration (TS.tv_nsec) / 10#1#E9;
   end To_Duration;

   function To_Duration (TV : struct_timeval) return Duration is
   begin
      return Duration (TV.tv_sec) + Duration (TV.tv_usec) / 10#1#E6;
   end To_Duration;

   -----------------
   -- To_Timespec --
   -----------------

   function To_Timespec (D : Duration) return timespec is
      S : time_t;
      F : Duration;

   begin
      S := time_t (Long_Long_Integer (D));
      F := D - Duration (S);

      --  If F has negative value due to a round-up, adjust for positive F
      --  value.

      if F < 0.0 then
         S := S - 1;
         F := F + 1.0;
      end if;

      return timespec' (tv_sec => S,
        tv_nsec => long (Long_Long_Integer (F * 10#1#E9)));
   end To_Timespec;

   ----------------
   -- To_Timeval --
   ----------------

   function To_Timeval (D : Duration) return struct_timeval is
      S : long;
      F : Duration;

   begin
      S := long (Long_Long_Integer (D));
      F := D - Duration (S);

      --  If F has negative value due to a round-up, adjust for positive F
      --  value.

      if F < 0.0 then
         S := S - 1;
         F := F + 1.0;
      end if;

      return struct_timeval' (tv_sec => S,
        tv_usec => long (Long_Long_Integer (F * 10#1#E6)));
   end To_Timeval;

   -------------------
   -- clock_gettime --
   -------------------

   function clock_gettime
     (clock_id : clockid_t;
      tp       : access timespec) return int
   is
      Result : int;
      tv : aliased struct_timeval;
   begin
      Result := gettimeofday (tv'Unchecked_Access, null);
      tp.all := To_Timespec (To_Duration (tv));
      return Result;
   end clock_gettime;

   ---------------------------
   --  POSIX.1c  Section 3  --
   ---------------------------

   function sigwait (set : access sigset_t; sig : access Signal) return int is
      Result : int;

      function sigwait (set : access sigset_t) return int;
      pragma Import (C, sigwait, "sigwait");

   begin
      Result := sigwait (set);

      if Result < 0 then
         sig.all := 0;
         return errno;
      end if;

      sig.all := Signal (Result);
      return 0;
   end sigwait;

   function pthread_kill (thread : pthread_t; sig : Signal) return int is
      function pthread_kill_base
        (thread : access pthread_t; sig : access Signal) return int;
      pragma Import (C, pthread_kill_base, "pthread_kill");

      thr   : aliased pthread_t := thread;
      signo : aliased Signal := sig;

   begin
      return pthread_kill_base (thr'Unchecked_Access, signo'Unchecked_Access);
   end pthread_kill;

   ----------------------------
   --  POSIX.1c  Section 13  --
   ----------------------------

   function pthread_attr_setschedparam
     (attr        : access pthread_attr_t;
      sched_param : int) return int
   is
      function pthread_attr_setschedparam_base
        (attr        : access pthread_attr_t;
         sched_param : access struct_sched_param) return int;
      pragma Import (C, pthread_attr_setschedparam_base,
        "pthread_attr_setschedparam");

      param : aliased struct_sched_param;

   begin
      param.sched_priority := sched_param;
      return pthread_attr_setschedparam_base (attr, param'Unchecked_Access);
   end pthread_attr_setschedparam;

   function pthread_attr_getschedparam
     (attr        : access pthread_attr_t;
      sched_param : access int) return int
   is
      function pthread_attr_getschedparam_base
        (attr        : access pthread_attr_t;
         sched_param : access struct_sched_param) return int;
      pragma Import (C, pthread_attr_getschedparam_base,
        "pthread_attr_getschedparam");

      param  : aliased struct_sched_param;
      Result : int;

   begin
      Result := pthread_attr_getschedparam_base (attr, param'Unchecked_Access);
      sched_param.all := param.sched_priority;
      return Result;
   end pthread_attr_getschedparam;

   function Get_Stack_Base (thread : pthread_t) return Address is
   begin
      return Null_Address;
   end Get_Stack_Base;

   procedure pthread_init is
   begin
      null;
   end pthread_init;

end System.OS_Interface;
